"
 COPYRIGHT (c) 1993 by Claus Gittinger
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

Object subclass:#WindowSensor
         instanceVariableNames:'eventSemaphore damage mouseAndKeyboard 
                                buttonState exposeEventSemaphore 
                                catchExpose gotExpose
                                compressMotionEvents'
         classVariableNames:''
         poolDictionaries:''
         category:'Interface-Support'
!

WindowSensor comment:'
Instances of this class keep track of events and damage areas for a group of views. 
All incoming expose rectangles and events (from Workstation) are collected here, until someone 
gets a chance to handle them. 
When adding an expose rectangle, WindowSensor tries to merge the rectangle with the list of 
existing damages to minimize redrawing.
'!

!WindowSensor class methodsFor:'instance creation'!

new
    ^ self basicNew initialize
! !

!WindowSensor methodsFor:'damage'!

addDamage: aRectangle view:aView
    "Add aRectangle to the damage list.
     We try to merge incoming rectangles with the existing damage by discarding
     incoming rectangles which are contained in the existing damage or replacing
     any existing damage rectangle with incoming rectangles that completely contain it.
     There could be much more optimization here (for example joining rectangles)..."

    | count temp index |

    damage size == 0 ifTrue: [
        damage := OrderedCollection with:(Array with:aView with:aRectangle)
    ] ifFalse: [
        "first look, if this rectangle is already in the expose list"
        damage do: [:d | 
            (d at:1) == aView ifTrue:[
                ((d at:2) contains: aRectangle) ifTrue: [^self]
            ].
        ].
        "then look, if new rectangle contains any in the expose list"
        count := 0.
        1 to: damage size do:[:i |
            |d|

            d := damage at:i.
            (d at:1) == aView ifTrue:[
                (aRectangle contains: (d at:2)) ifTrue: [ 
                    damage at:i put: nil.
                    count := count + 1
                ]
            ]
        ].
        count == 0 ifTrue: [
            damage add: (Array with:aView with:aRectangle)
        ] ifFalse: [
            temp := OrderedCollection new: (damage size - count + 1).
            index := 1.
            damage do:[:aDamage |
                aDamage notNil ifTrue: [
                    temp add: aDamage.
                ]
            ].
            temp add: (Array with:aView with:aRectangle).
            damage := temp
        ]
    ].
!

nextDamage
    "retrieve the next damage rectangle or nil, if there is none"

    damage size == 0 ifTrue:[
        ^ nil
    ].
    ^ damage removeFirst.
!

nextEvent
    "retrieve the next event or nil, if there is none"

    mouseAndKeyboard size == 0 ifTrue:[
        ^ nil
    ].
    ^ mouseAndKeyboard removeFirst.
! !

!WindowSensor methodsFor:'special'!

suspendForExpose
    "asynchronous wait until a no expose arrives (after a scroll)"

    exposeEventSemaphore := Semaphore new.
    exposeEventSemaphore wait.
    exposeEventSemaphore := nil
!

catchExpose
    "start catching noExpose events (must be done after the bitblt,
     if some time may pass till the noExpose is been waited for."

    catchExpose := true.
    gotExpose := false.
!

waitForExposeFor:aView
    "synchronous wait until a no expose arrives (after a scroll)"

" "
    aView device synchronizeOutput.
" "
    [gotExpose] whileFalse:[
        aView device dispatchExposeEventFor:aView id
    ].
    catchExpose := false
! !

!WindowSensor methodsFor:'event processing'!

notifyEventArrival
    "an event arrived - if there is an eventsemaphore,
     signal it, to wake up any controller process"

    catchExpose == true ifTrue:[
        "dont wake up, if we are currently waiting for an expose"
        ^ self
    ].
    eventSemaphore notNil ifTrue:[
        eventSemaphore signal
    ]
!

hasDamage 
    "return true, if any exposure events are pending"

    ^ damage size ~~ 0
!

hasEvents 
    "return true, if any mouse/keyboard events are pending"

    ^ mouseAndKeyboard size ~~ 0
!

eventPending
    "return true, if either damage or events are pending"

    mouseAndKeyboard size ~~ 0 ifTrue:[^ true].
    ^ damage size ~~ 0
!

exposeX:left y:top width:width height:height view:aView
    self addDamage:(left @ top extent:width @ height) view:aView.
    self notifyEventArrival
!

noExposeView:aView
    gotExpose := true.
    catchExpose := false.

    exposeEventSemaphore notNil ifTrue:[
        exposeEventSemaphore signal
    ]
!

graphicExposeX:left y:top width:width height:height view:aView
"
'g' print.
"
    "do it now ..."
" 
    aView exposeX:left y:top width:width height:height.
    aView device synchronizeOutput
" 

    "... or better queue it ?"
" "
    self addDamage:(left @ top extent:width @ height) view:aView.
" "
!

buttonMotion:state x:x y:y view:aView
    "mouse was moved"

    compressMotionEvents ifTrue:[
        mouseAndKeyboard reverseDo:[:ev |
            ((ev type == #buttonMotion:x:y:) 
            and:[(ev view == aView)
            and:[(ev arguments at:1) == state]]) ifTrue:[
                ev arguments:(Array with:state with:x with:y).
                ^ self
            ]
        ]
    ].
"
'M' print.
"
    mouseAndKeyboard addLast:(WindowEvent for:aView type:#buttonMotion:x:y:
                                               arguments:(Array with:state
                                                                with:x
                                                                with:y)).
    self notifyEventArrival
!

buttonPress:button x:x y:y view:aView
    "mouse button was pressed"

    mouseAndKeyboard addLast:(WindowEvent for:aView type:#buttonPress:x:y:
                                               arguments:(Array with:button
                                                                with:x
                                                                with:y)).
    self notifyEventArrival
!

buttonRelease:button x:x y:y view:aView
    "mouse button was released"

    mouseAndKeyboard addLast:(WindowEvent for:aView type:#buttonRelease:x:y:
                                               arguments:(Array with:button
                                                                with:x
                                                                with:y)).
    self notifyEventArrival
!

pointerEnter:state x:x y:y view:aView
    "mouse cursor was moved into the view"

    mouseAndKeyboard addLast:(WindowEvent for:aView type:#pointerEnter:x:y:
                                               arguments:(Array with:state 
                                                                with:x
                                                                with:y)).
    self notifyEventArrival
!

pointerLeave:state view:aView
    "mouse cursor was moved out of the view"

    mouseAndKeyboard addLast:(WindowEvent for:aView type:#pointerLeave: 
                                               arguments:(Array with:state)).
    self notifyEventArrival
!

buttonMultiPress:button x:x y:y view:aView
    "mouse button was pressed"

    mouseAndKeyboard addLast:(WindowEvent for:aView type:#buttonMultiPress:x:y:
                                               arguments:(Array with:button
                                                                with:x
                                                                with:y)).
    self notifyEventArrival
!

buttonShiftPress:button x:x y:y view:aView
    "mouse button was pressed"

    mouseAndKeyboard addLast:(WindowEvent for:aView type:#buttonShiftPress:x:y:
                                               arguments:(Array with:button
                                                                with:x
                                                                with:y)).
    self notifyEventArrival
!

keyPress:key x:x y:y view:aView
    "key was pressed"

    mouseAndKeyboard addLast:(WindowEvent for:aView type:#keyPress:x:y:
                                               arguments:(Array with:key 
                                                                with:x
                                                                with:y)).
    self notifyEventArrival
! !

!WindowSensor methodsFor:'initialization'!

initialize
    "initialize the event queues to empty"

    damage := OrderedCollection new.
    mouseAndKeyboard := OrderedCollection new.
    gotExpose := true.
    catchExpose := false.
    compressMotionEvents := true
! !

!WindowSensor methodsFor:'accessing'!

eventSemaphore:aSemaphore
    "set the semaphore used to signal event arrival"

    eventSemaphore := aSemaphore
!

eventSemaphore
    "return the semaphore used to signal event arrival"

    ^ eventSemaphore
! !
